# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

from pathlib import Path
from unittest.mock import patch

import pytest
from antsibull_core.yaml import load_yaml_file

from antsibull.cli.antsibull_build import run


@pytest.mark.parametrize(
    "args, expected, ret",
    [
        pytest.param(
            [],
            [
                "cisco.nso 1.0.3 is not tagged in https://github.com/CiscoDevNet/ansible-nso",
                "hpe.nimble 1.1.4 is not tagged in https://github.com/hpe-storage/nimble-ansible-modules",
                "inspur.sm 2.3.0 is not tagged in https://github.com/ISIB-Group/inspur.sm",
                "mellanox.onyx 1.0.0 is not tagged in https://github.com/ansible-collections/mellanox.onyx",
            ],
            1,
            id="simple",
        ),
        pytest.param(
            ["-I", "cisco.nso"],
            [
                "hpe.nimble 1.1.4 is not tagged in https://github.com/hpe-storage/nimble-ansible-modules",
                "inspur.sm 2.3.0 is not tagged in https://github.com/ISIB-Group/inspur.sm",
                "mellanox.onyx 1.0.0 is not tagged in https://github.com/ansible-collections/mellanox.onyx",
            ],
            1,
            id="one-ignore",
        ),
        pytest.param(
            ["-I", "cisco.nso", "-I", "xyz"],
            [
                "hpe.nimble 1.1.4 is not tagged in https://github.com/hpe-storage/nimble-ansible-modules",
                "inspur.sm 2.3.0 is not tagged in https://github.com/ISIB-Group/inspur.sm",
                "mellanox.onyx 1.0.0 is not tagged in https://github.com/ansible-collections/mellanox.onyx",
                "invalid ignore 'xyz': xyz does not match any collection",
            ],
            1,
            id="one-ignore-with-invalid",
        ),
        pytest.param(
            [
                "-I",
                "cisco.nso",
                "-I",
                "hpe.nimble",
                "-I",
                "inspur.sm",
                "-I",
                "mellanox.onyx",
            ],
            [],
            0,
            id="ignore-all",
        ),
        pytest.param(
            [
                "-I",
                "cisco.nso",
                "-I",
                "hpe.nimble",
                "-I",
                "inspur.sm",
                "-I",
                "mellanox.onyx",
                "-I",
                "asdf",
                "-I",
                "community.general",
            ],
            [
                "invalid ignore 'asdf': asdf does not match any collection",
                "useless ignore 'community.general':"
                " community.general 6.5.0 is properly tagged",
            ],
            1,
            id="ignore-all-with-invalid",
        ),
    ],
)
def test_validate_tags_file(
    test_data_path: Path,
    capsys: pytest.CaptureFixture,
    args: list[str],
    expected: list[str],
    ret: int,
):
    path = test_data_path / "ansible-7.4.0-tags.yaml"
    assert run(["antsibull-build", "validate-tags-file", str(path), *args]) == ret
    out, err = capsys.readouterr()
    assert sorted(err.splitlines()) == sorted(expected)


@pytest.mark.parametrize(
    "args, ignore_file_contents, expected, ret",
    [
        pytest.param(
            [],
            [],
            [
                "cisco.nso 1.0.3 is not tagged in https://github.com/CiscoDevNet/ansible-nso",
                "hpe.nimble 1.1.4 is not tagged in https://github.com/hpe-storage/nimble-ansible-modules",
                "inspur.sm 2.3.0 is not tagged in https://github.com/ISIB-Group/inspur.sm",
                "mellanox.onyx 1.0.0 is not tagged in https://github.com/ansible-collections/mellanox.onyx",
            ],
            1,
            id="simple",
        ),
        pytest.param(
            [],
            ["cisco.nso"],
            [
                "hpe.nimble 1.1.4 is not tagged in https://github.com/hpe-storage/nimble-ansible-modules",
                "inspur.sm 2.3.0 is not tagged in https://github.com/ISIB-Group/inspur.sm",
                "mellanox.onyx 1.0.0 is not tagged in https://github.com/ansible-collections/mellanox.onyx",
            ],
            1,
            id="one-ignore",
        ),
        pytest.param(
            [],
            ["cisco.nso", "xyz"],
            [
                "hpe.nimble 1.1.4 is not tagged in https://github.com/hpe-storage/nimble-ansible-modules",
                "inspur.sm 2.3.0 is not tagged in https://github.com/ISIB-Group/inspur.sm",
                "mellanox.onyx 1.0.0 is not tagged in https://github.com/ansible-collections/mellanox.onyx",
                "invalid ignore 'xyz': xyz does not match any collection",
            ],
            1,
            id="one-ignore-with-invalid",
        ),
        pytest.param(
            [],
            [
                "cisco.nso",
                "hpe.nimble",
                "inspur.sm",
                "mellanox.onyx",
            ],
            [],
            0,
            id="ignore-all",
        ),
        pytest.param(
            [],
            [
                "cisco.nso",
                "hpe.nimble",
                "inspur.sm",
                "mellanox.onyx",
                "asdf",
            ],
            ["invalid ignore 'asdf': asdf does not match any collection"],
            1,
            id="ignore-all-with-invalid",
        ),
        pytest.param(
            ["-I", "mellanox.onyx", "-I", "xyz"],
            [
                "cisco.nso",
                "hpe.nimble",
                "inspur.sm",
                "asdf",
            ],
            [
                "invalid ignore 'xyz': xyz does not match any collection",
                "invalid ignore 'asdf': asdf does not match any collection",
            ],
            1,
            id="mixed",
        ),
    ],
)
def test_validate_tags_file_ignore_file(
    test_data_path: Path,
    tmp_path: Path,
    capsys: pytest.CaptureFixture,
    args: list[str],
    ignore_file_contents: list[str],
    expected: list[str],
    ret: int,
):
    path = test_data_path / "ansible-7.4.0-tags.yaml"
    ignores_file = tmp_path / "ignores_file"
    ignores_file.write_text("\n".join(ignore_file_contents))
    ran = run(
        [
            "antsibull-build",
            "validate-tags-file",
            str(path),
            "--ignores-file",
            str(ignores_file),
            *args,
        ]
    )
    assert ran == ret
    out, err = capsys.readouterr()
    assert sorted(err.splitlines()) == sorted(expected)


def test_validate_tags(test_data_path: Path, tmp_path: Path):
    ignores_file = test_data_path / "validate-tags-ignores"
    name = "ansible-7.4.0-tags.yaml"
    expected_data_path = test_data_path / name
    expected_data = load_yaml_file(expected_data_path)
    output_data_path = tmp_path / name
    with patch("antsibull.tagging.get_collections_tags", return_value=expected_data):
        ran = run(
            [
                "antsibull-build",
                "validate-tags",
                f"--data-dir={test_data_path}",
                f"--ignores-file={ignores_file}",
                f"--output={output_data_path}",
                "--error-on-useless-ignores",
                "7.4.0",
            ]
        )
    assert ran == 0
    output_data = load_yaml_file(output_data_path)
    assert expected_data == output_data



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+
# (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Work with the PyPI API
"""

import datetime
from pathlib import Path
from typing import NamedTuple, Optional
from urllib.parse import urljoin

import aiohttp
import pydantic as p
from antsibull_core.utils.hashing import verify_hash

PYPI_BASE_URL = "https://pypi.org/pypi/"


class SdistAndWheelPair(NamedTuple):
    """
    Tuple of UrlInfo objects for a project's sdist and wheel
    """

    sdist: "UrlInfo"
    wheel: "UrlInfo"


# pyre-ignore[13]: BaseModel initializes attributes when data is loaded
class Release(p.BaseModel):
    """
    Model representing the response from the PyPI API releases endpoint.

    https://warehouse.pypa.io/api-reference/json.html#get--pypi--project_name---version--json
    """

    info: "ReleaseInfo"
    last_serial: int
    urls: list["UrlInfo"]
    vulnerabilities: list[dict]

    def get_sdist_and_wheel(self) -> SdistAndWheelPair:
        sdist: Optional[UrlInfo] = None
        wheel: Optional[UrlInfo] = None
        for release in self.urls:
            if not sdist and release.packagetype == "sdist":
                sdist = release
            elif not wheel and release.packagetype == "bdist_wheel":
                wheel = release
            else:
                break
            msg = f"Not {{}} was uploaded for {self.info.name}=={self.info.version}"
        if not sdist:
            raise ValueError(msg.format("sdist"))
        if not wheel:
            raise ValueError(msg.format("wheel"))
        return SdistAndWheelPair(sdist, wheel)


# pyre-ignore[13]: BaseModel initializes attributes when data is loaded
class ReleaseInfo(p.BaseModel):
    author: str
    author_email: str
    classifiers: list[str]
    description: str
    description_content_type: str
    docs_url: Optional[str]
    home_page: str
    keywords: str
    license: str
    maintainer: str
    maintainer_email: str
    name: str
    package_url: str
    platform: Optional[str]
    project_url: str
    project_urls: dict[str, str]
    release_url: str
    requires_dist: list[str]
    requires_python: str
    summary: str
    version: str
    yanked: bool
    yanked_reason: Optional[str]


# pyre-ignore[13]: BaseModel initializes attributes when data is loaded
class UrlInfo(p.BaseModel):
    comment_text: str
    digests: dict[str, str]
    filename: str
    md5_digest: str
    packagetype: str
    python_version: str
    requires_python: Optional[str]
    size: int
    upload_time: datetime.datetime
    url: str
    yanked: bool
    yanked_reason: Optional[str]

    @property
    def sha256sum(self) -> str:
        return self.digests["sha256"]

    async def verify_local_file(self, file: Path) -> bool:
        """
        Check if a local file's name and sha256sum matches this release
        """
        if file.name != self.filename:
            return False
        return await verify_hash(file, self.sha256sum)


# TODO pydantic v2+:
# + Release.model_rebuild
getattr(Release, "model_rebuild", getattr(Release, "update_forward_refs"))()


class PyPIClient:
    """
    Client for the PyPI Warehouse JSON API
    """

    def __init__(self, aio_session: aiohttp.ClientSession) -> None:
        self.aio_session = aio_session

    async def get_release(self, package: str, version: str) -> Release:
        url = urljoin(PYPI_BASE_URL, f"{package}/{version}/json")
        async with self.aio_session.get(url) as resp:
            resp.raise_for_status()
            data = await resp.json()
        return Release(**data)


__all__ = (
    "SdistAndWheelPair",
    "Release",
    "ReleaseInfo",
    "UrlInfo",
    "PyPIClient",
)



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+
# (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Commands for generating release announcements
"""

from __future__ import annotations

import asyncio
import json
import sys
import webbrowser
from collections.abc import Callable, Iterator
from dataclasses import dataclass
from functools import lru_cache, partial
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING
from urllib.parse import quote as url_quote
from urllib.parse import urlencode, urljoin

import pydantic.json
from aiohttp import ClientResponseError, ClientSession
from antsibull_core import app_context
from antsibull_core.dependency_files import DependencyFileData, DepsFile
from jinja2 import Environment, PackageLoader, StrictUndefined, select_autoescape
from packaging.version import Version as PypiVer
from typing_extensions import TypedDict

from antsibull.constants import ANSIBLE_FORUM_URL, BUILD_DATA_URL
from antsibull.pypi import PyPIClient, SdistAndWheelPair, UrlInfo

try:
    import pyperclip  # type: ignore[import]
except ImportError:
    HAS_PYPERCLIP = False
else:
    HAS_PYPERCLIP = True

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath

ANNOUNCEMENTS = {
    "ansible-email-announcement.txt": "ansible-email-announcement.j2",
    "ansible-matrix-announcement.md": "ansible-matrix-announcement.j2",
}
SUBJECT = "Release announcement: Ansible community package {version}"
SUBJECT_PRE_RELEASE = SUBJECT + " (Pre-Release)"
MAIL_RECIPIENTS = (
    "ansible-devel@googlegroups.com",
    "ansible-project@googlegroups.com",
    "ansible-announce@googlegroups.com",
)
# pylint: disable-next=line-too-long
# https://meta.discourse.org/t/create-a-link-to-start-a-new-topic-with-pre-filled-information/28074 # noqa
FORUM_TAGS = ("release", "distro-packaging", "release-management")
FORUM_PARAMS: dict[str, str] = {
    "category": "news/releases",
    "tags": ",".join(FORUM_TAGS),
}

eprint = partial(print, file=sys.stderr)

jinja_env = Environment(
    loader=PackageLoader(__package__, "data"),
    autoescape=select_autoescape(),
    trim_blocks=True,
    undefined=StrictUndefined,
)


def email_heading(content):
    """
    Given a string, convert it into an email heading

    Args:
        content: filter content
    """
    return content + "\n" + "-" * len(content)


jinja_env.filters["email_heading"] = email_heading


class TemplateVars(TypedDict):
    """
    Variables used in the Jinja2 template
    """

    version: str
    major_version: int
    core_version: str
    core_major_version: str
    build_data_path: str
    release_tarball: UrlInfo
    release_wheel: UrlInfo
    is_major_release: bool
    is_prerelease: bool


# pyre-ignore[13]: BaseModel initializes attributes when data is loaded
class AnnouncementsInfo(pydantic.BaseModel):
    """
    Represents the announcement files generated by `announcements_command`
    """

    template_vars: TemplateVars
    outputs: list[str]


def announcements_command() -> int:
    """
    Generate release announcements
    """
    app_ctx = app_context.app_ctx.get()
    ansible_version: str = app_ctx.extra["ansible_version"]
    output_dir: Path = app_ctx.extra["output_dir"]
    dist_dir: Path | None = app_ctx.extra["dist_dir"]
    send: bool = app_ctx.extra["send"]

    deps_filename = Path(app_ctx.extra["data_dir"], app_ctx.extra["deps_file"])
    deps_file = DepsFile(deps_filename)
    dependency_data = deps_file.parse()
    return asyncio.run(
        _announcements_command(
            ansible_version, output_dir, dist_dir, dependency_data, send
        )
    )


async def verify_dists(dists: SdistAndWheelPair, dist_dir: Path) -> str | None:
    """
    Verify an SdistAndWheelPair to ensure it matches files on the local filesystem

    Args:
        dists: `SdistAndWheelPair` object
        dist_dir: Directory with appropriately named dist files to validate

    Returns:
        Returns an error message as a `str` or `None` if no errors were found
    """
    for dist in dists:
        dist_path = dist_dir / dist.filename
        if not await asyncio.to_thread(dist_path.is_file):
            return f"{dist_path.name} was not found in --dist-dir"
        if not await dist.verify_local_file(dist_path):
            return f"{dist_path} differs from {dist.url}"
    return None


def write_announcements(
    announcements: dict[str, str], ctx: TemplateVars, output_dir: Path
) -> Iterator[Path]:
    """
    Write out announcement templates to `output_dir`

    Args:
        announcements:
            Mapping of output filename to the filename of the Jinja2 template file
        ctx:
            A `TemplateVars` dictionary
        output_dir:
            Directory in which to output the template files

    Yields:
        `pathlib.Path` objects of the output files
    """
    for name, template in announcements.items():
        output = jinja_env.get_template(template).render(**ctx)
        path = output_dir.joinpath(name)
        path.write_text(output + "\n")
        yield path


async def get_data(
    ansible_version: str, dist_dir: Path | None, dependency_data: DependencyFileData
) -> TemplateVars | None:
    """
    Retrieve package data from PyPI and return a `TemplateVars` dictionary

    Args:
        ansible_version:
            Version of the `ansible` package to validate
        dist_dir:
            See `verify_dists()`. Set to `None` to disable validation of local
            dist files.
        dependency_data:
            `DependencyFileData` object
    """
    async with ClientSession() as aio_session:
        client = PyPIClient(aio_session)
        try:
            release = await client.get_release("ansible", ansible_version)
        except ClientResponseError as exc:
            eprint(f"Failed to retrieve data for ansible=={ansible_version}: {exc}")
            return None
        try:
            dists = release.get_sdist_and_wheel()
        except ValueError as exc:
            eprint(exc)
            return None
        if dist_dir and (err := await verify_dists(dists, dist_dir)):
            eprint(err)
            return None

        version = dependency_data.ansible_version
        version_obj = PypiVer(version)
        major_version = version_obj.major
        core_version = dependency_data.ansible_core_version
        core_version_obj = PypiVer(dependency_data.ansible_core_version)
        core_major_version = f"{core_version_obj.major}.{core_version_obj.minor}"
        build_data_path = f"{BUILD_DATA_URL}/blob/{version}/{major_version}"
        ctx = TemplateVars(
            version=version,
            major_version=major_version,
            core_version=core_version,
            core_major_version=core_major_version,
            build_data_path=build_data_path,
            release_tarball=dists.sdist,
            release_wheel=dists.wheel,
            is_major_release=version_obj.minor == 0 and version_obj.micro == 0,
            is_prerelease=version_obj.pre is not None,
        )
        return ctx


async def _announcements_command(
    ansible_version: str,
    output_dir: Path,
    dist_dir: Path | None,
    dependency_data: DependencyFileData,
    send: bool,
) -> int:
    if not (ctx := await get_data(ansible_version, dist_dir, dependency_data)):
        return 1
    for path in write_announcements(ANNOUNCEMENTS, ctx, output_dir):
        print("Wrote:", path)
    data = AnnouncementsInfo(template_vars=ctx, outputs=list(ANNOUNCEMENTS))
    info_path = output_dir / "announcements.json"
    write_announcements_json(data, info_path)
    print("Wrote:", info_path)
    if send and (rc := _send_announcements_command(output_dir, set(ACTIONS), True)):
        return rc
    return 0


def write_announcements_json(info: AnnouncementsInfo, file: StrOrBytesPath) -> None:
    """
    Write the announcements.json data file to the filesystem

    Args:
        info: `AnnouncementsInfo` object
        file: Output JSON filename
    """
    with open(file, "w", encoding="utf-8") as fp:
        # pyre and pylint don't understand pydantic's __getattr__ magic
        # pyre-ignore[16] pylint: disable-next=no-member
        json.dump(info, fp, default=pydantic.json.pydantic_encoder, indent=2)


def load_announcements_json(file: StrOrBytesPath) -> AnnouncementsInfo:
    """
    Load the announcement.json data file as an AnnouncementsInfo object

    Args:
        file: JSON filename
    """
    with open(file, "r", encoding="utf-8") as fp:
        data = json.load(fp)
    return AnnouncementsInfo(**data)


@lru_cache
def get_body(directory: Path, name: str) -> str:
    """
    Memoized function to retrieve a text file named `name` from `directory`.

    Args:
        directory: directory in which the file resides
        name: basename of the file
    """
    return (directory / name).read_text()


def get_subject(info: AnnouncementsInfo) -> str:
    template = SUBJECT_PRE_RELEASE if info.template_vars["is_prerelease"] else SUBJECT
    return template.format(version=info.template_vars["version"])


def forum_announcement_webbrowser(
    directory: Path,
    info: AnnouncementsInfo,
    ctx: SendCtx,  # pylint: disable=unused-argument
) -> None:
    """
    Open a pre-filled Ansible Forum post in a browser
    """
    subject = get_subject(info)
    body = get_body(directory, "ansible-email-announcement.txt")
    params_dict = FORUM_PARAMS | {"title": subject, "body": body}
    params = "?" + urlencode(params_dict, quote_via=url_quote)
    url = urljoin(ANSIBLE_FORUM_URL, "new-topic") + params
    webbrowser.open(url)


def email_announcement_webbrowser(
    directory: Path,
    info: AnnouncementsInfo,
    ctx: SendCtx,  # pylint: disable=unused-argument
) -> None:
    """
    Construct a pre-filled mailto link with the appropriate subject, body,
    and recipients and open it an a browser
    """
    subject = get_subject(info)
    body = get_body(directory, "ansible-email-announcement.txt")
    params_dict: dict[str, str] = {"subject": subject, "body": body}
    params = "?" + urlencode(params_dict, quote_via=url_quote)
    url = "mailto:" + ",".join(MAIL_RECIPIENTS) + params
    webbrowser.open(url)


def matrix_announcement(
    directory: Path,
    info: AnnouncementsInfo,  # pylint: disable=unused-argument
    ctx: SendCtx,
) -> None:
    """
    Prompt user to send a Matrix release announcement
    """
    body = get_body(directory, "ansible-matrix-announcement.md")
    try:
        forum_url = input("Enter the URL to the forum post: ")
    # Ctrl+D
    except EOFError:
        print("Continuing...")
        return
    body = body.replace("<FORUM LINK>", forum_url)
    message = dedent(
        """
    Please open your Matrix client and send the message to:
    - #community:ansible.com
    - #packaging:ansible.com
    - #social:ansible.com (mention @newsbot)
    """
    )
    if ctx.clipboard:
        pyperclip.copy(body)
        message += "\nThe message has been copied to your clipboard"
    else:
        message += "---\n" + body
    print(message)


ACTIONS: dict[str, Callable[[Path, AnnouncementsInfo, SendCtx], None]] = {
    "forum": forum_announcement_webbrowser,
    "email": email_announcement_webbrowser,
    "matrix": matrix_announcement,
}


@dataclass
class SendCtx:
    """
    Context with options to be used by the `*_announcement_*` functions
    """

    # Whether actions can write to the clipboard
    clipboard: bool


def send_announcements_command() -> int:
    """
    Interactively send announcements created by the `announcements` subcommand.
    """
    app_ctx = app_context.app_ctx.get()
    announcements_dir: Path = app_ctx.extra["announcements_dir"]
    actions: set[str] = set(app_ctx.extra["send_actions"])
    clipboard: bool = app_ctx.extra["clipboard"]

    if clipboard and not HAS_PYPERCLIP:
        eprint(
            "The pyperclip library is not installed."
            " Install it with 'pip install antsibull[clipboard]' or use '--no-clipboard'."
        )
        return 1
    return _send_announcements_command(announcements_dir, actions, clipboard)


def _send_announcements_command(
    announcements_dir: Path, actions: set[str], clipboard: bool
) -> int:
    info = load_announcements_json(announcements_dir / "announcements.json")
    send_ctx = SendCtx(clipboard)
    # Loop over ACTIONS this way to make sure that order is preserved.
    # For example, we want the Matrix announcement to happen last
    for action, func in ACTIONS.items():
        if action not in actions:
            continue
        print(f"Handling {action}...")
        func(announcements_dir, info, send_ctx)

    return 0



# Author: Felix Fontein <felix@fontein.de>
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020
"""Changelog handling and processing code."""

from __future__ import annotations

import asyncio
import datetime
import glob
import os
import os.path
import tarfile
import tempfile
import typing as t
from collections import defaultdict

import aiohttp
import asyncio_pool  # type: ignore[import]
from antsibull_changelog.changes import ChangesData, add_release
from antsibull_changelog.config import ChangelogConfig, CollectionDetails, PathsConfig
from antsibull_changelog.rendering.changelog import ChangelogGenerator
from antsibull_changelog.utils import collect_versions
from antsibull_core import app_context
from antsibull_core.ansible_core import get_ansible_core
from antsibull_core.dependency_files import DependencyFileData, DepsFile
from antsibull_core.galaxy import CollectionDownloader, GalaxyContext
from antsibull_core.yaml import load_yaml_bytes
from packaging.version import Version as PypiVer
from semantic_version import Version as SemVer

from antsibull.collection_meta import CollectionsMetadata


class ChangelogData:
    """
    Data for a single changelog (for a collection, for ansible-core, for Ansible)
    """

    paths: PathsConfig
    config: ChangelogConfig
    changes: ChangesData
    generator: ChangelogGenerator
    generator_flatmap: bool

    def __init__(
        self,
        paths: PathsConfig,
        config: ChangelogConfig,
        changes: ChangesData,
        flatmap: bool = False,
    ):
        self.paths = paths
        self.config = config
        self.changes = changes
        self.generator_flatmap = flatmap
        self.generator = ChangelogGenerator(
            self.config, self.changes, plugins=None, fragments=None, flatmap=flatmap
        )

    @classmethod
    def collection(
        cls, collection_name: str, version: str, changelog_data: t.Any | None = None
    ) -> ChangelogData:
        paths = PathsConfig.force_collection("")
        collection_details = CollectionDetails(paths)
        collection_details.namespace, collection_details.name = collection_name.split(
            ".", 1
        )
        collection_details.version = version
        collection_details.flatmap = False  # TODO!
        config = ChangelogConfig.default(paths, collection_details)
        return cls(
            paths, config, ChangesData(config, "", changelog_data), flatmap=True
        )  # TODO!

    @classmethod
    def ansible_core(cls, changelog_data: t.Any | None = None) -> ChangelogData:
        paths = PathsConfig.force_ansible("")
        collection_details = CollectionDetails(paths)
        config = ChangelogConfig.default(paths, collection_details)
        return cls(
            paths, config, ChangesData(config, "", changelog_data), flatmap=False
        )

    @classmethod
    def ansible(
        cls, directory: str | None, output_directory: str | None = None
    ) -> ChangelogData:
        paths = PathsConfig.force_ansible("")

        config = ChangelogConfig.default(paths, CollectionDetails(paths), "Ansible")
        # TODO: adjust the following lines once Ansible switches to semantic versioning
        config.use_semantic_versioning = False
        config.release_tag_re = r"""(v(?:[\d.ab\-]|rc)+)"""
        config.pre_release_tag_re = r"""(?P<pre_release>(?:[ab]|rc)+\d*)$"""

        changelog_path = ""
        if directory is not None:
            changelog_path = os.path.join(directory, "changelog.yaml")
        changes = ChangesData(config, changelog_path)
        if output_directory is not None:
            changes.path = os.path.join(output_directory, "changelog.yaml")
        return cls(paths, config, changes, flatmap=True)

    @classmethod
    def concatenate(cls, changelogs: list[ChangelogData]) -> ChangelogData:
        return cls(
            changelogs[0].paths,
            changelogs[0].config,
            ChangesData.concatenate([changelog.changes for changelog in changelogs]),
            flatmap=changelogs[0].generator_flatmap,
        )

    def add_ansible_release(
        self, version: str, date: datetime.date, release_summary: str
    ) -> None:
        add_release(
            self.config,
            self.changes,
            [],
            [],
            version,
            codename=None,
            date=date,
            update_existing=True,
            show_release_summary_warning=False,
        )
        release_date = self.changes.releases[version]
        if "changes" not in release_date:
            release_date["changes"] = {}
        release_date["changes"]["release_summary"] = release_summary


def read_file(tarball_path: str, matcher: t.Callable[[str], bool]) -> bytes | None:
    with tarfile.open(tarball_path, "r:gz") as tar:
        for file in tar:
            if matcher(file.name):
                file_p = tar.extractfile(file)
                if file_p:
                    with file_p:
                        return file_p.read()
    return None


def read_changelog_file(tarball_path: str, is_ansible_core=False) -> bytes | None:
    def matcher(filename: str) -> bool:
        if is_ansible_core:
            return filename.endswith("changelogs/changelog.yaml")
        return filename in ("changelogs/changelog.yaml", "changelog.yaml")

    return read_file(tarball_path, matcher)


def get_porting_guide_filename(version: PypiVer):
    return f"docs/docsite/rst/porting_guides/porting_guide_core_{version.major}.{version.minor}.rst"


class CollectionChangelogCollector:
    collection: str
    versions: list[SemVer]
    earliest: SemVer
    latest: SemVer

    changelog: ChangelogData | None

    def __init__(self, collection: str, versions: t.ValuesView[str]):
        self.collection = collection
        self.versions = sorted(SemVer(version) for version in versions)
        self.earliest = self.versions[0]
        self.latest = self.versions[-1]
        self.changelog = None

    async def _get_changelog(
        self, version: SemVer, collection_downloader: CollectionDownloader
    ) -> ChangelogData | None:
        path = await collection_downloader.download(self.collection, version)
        changelog_bytes = read_changelog_file(path)
        if changelog_bytes is None:
            return None
        try:
            changelog_data = load_yaml_bytes(changelog_bytes)
            return ChangelogData.collection(
                self.collection, str(version), changelog_data
            )
        except Exception as exc:  # pylint: disable=broad-except
            print(
                f"WARNING: cannot load changelog of {self.collection} {version} due to {exc}"
            )
            return None

    async def _download_changelog_stream(
        self, start_version: SemVer, collection_downloader: CollectionDownloader
    ) -> ChangelogData | None:
        changelog = await self._get_changelog(start_version, collection_downloader)
        if changelog is None:
            return None

        changelog.changes.prune_versions(
            versions_after=None, versions_until=str(start_version)
        )
        changelogs = [changelog]
        ancestor = changelog.changes.ancestor
        while ancestor is not None:
            ancestor_ver = SemVer(ancestor)
            if ancestor_ver < self.earliest:
                break
            changelog = await self._get_changelog(ancestor_ver, collection_downloader)
            if changelog is None:
                break
            changelog.changes.prune_versions(
                versions_after=None, versions_until=ancestor
            )
            changelogs.append(changelog)
            ancestor = changelog.changes.ancestor

        return ChangelogData.concatenate(changelogs)

    async def download(self, collection_downloader: CollectionDownloader):
        missing_versions = set(self.versions)

        while missing_versions:
            missing_version = max(missing_versions)

            # Try to get hold of changelog for this version
            changelog = await self._download_changelog_stream(
                missing_version, collection_downloader
            )
            if changelog:
                current_changelog = self.changelog
                if current_changelog is None:
                    # If we didn't have a changelog so far, start with it
                    self.changelog = changelog
                    missing_versions -= {
                        SemVer(version) for version in changelog.changes.releases
                    }
                else:
                    # Insert entries from changelog into combined changelog that are missing there
                    for version, entry in changelog.changes.releases.items():
                        sem_version = SemVer(version)
                        if sem_version in missing_versions:
                            current_changelog.changes.releases[version] = entry
                            missing_versions.remove(sem_version)

            # Make sure that this version isn't checked again
            missing_versions -= {missing_version}


class AnsibleCoreChangelogCollector:
    versions: list[PypiVer]
    earliest: PypiVer
    latest: PypiVer

    changelog: ChangelogData | None

    porting_guide: bytes | None

    def __init__(self, versions: t.ValuesView[str]):
        self.versions = sorted(PypiVer(version) for version in versions)
        self.earliest = self.versions[0]
        self.latest = self.versions[-1]
        self.changelog = None
        self.porting_guide = None

    @staticmethod
    async def _get_changelog_file(
        version: PypiVer, core_downloader: t.Callable[[str], t.Awaitable[str]]
    ) -> ChangelogData | None:
        path = await core_downloader(str(version))
        if os.path.isdir(path):
            changelog: ChangelogData | None = None
            for root, dummy, files in os.walk(path):
                if "changelog.yaml" in files:
                    with open(os.path.join(root, "changelog.yaml"), "rb") as f:
                        changelog_bytes = f.read()
                    changelog_data = load_yaml_bytes(changelog_bytes)
                    changelog = ChangelogData.ansible_core(changelog_data)
            return changelog
        if os.path.isfile(path) and path.endswith(".tar.gz"):
            maybe_changelog_bytes = read_changelog_file(path, is_ansible_core=True)
            if maybe_changelog_bytes is None:
                return None
            changelog_data = load_yaml_bytes(maybe_changelog_bytes)
            return ChangelogData.ansible_core(changelog_data)
        return None

    async def download_changelog(
        self, core_downloader: t.Callable[[str], t.Awaitable[str]]
    ):
        changelog = await self._get_changelog_file(self.latest, core_downloader)
        if changelog is None:
            return

        changelog.changes.prune_versions(
            versions_after=None, versions_until=str(self.latest)
        )

        changelogs = [changelog]
        ancestor = changelog.changes.ancestor
        while ancestor is not None:
            ancestor_ver = PypiVer(ancestor)
            if ancestor_ver < self.earliest:
                break
            changelog = await self._get_changelog_file(ancestor_ver, core_downloader)
            if changelog is None:
                break
            changelog.changes.prune_versions(
                versions_after=None, versions_until=ancestor
            )
            changelogs.append(changelog)
            ancestor = changelog.changes.ancestor

        self.changelog = ChangelogData.concatenate(changelogs)

    async def download_porting_guide(self, aio_session: aiohttp.client.ClientSession):
        branch_url = (
            "https://raw.githubusercontent.com/ansible/ansible-documentation/devel"
        )

        query_url = f"{branch_url}/{get_porting_guide_filename(self.latest)}"
        async with aio_session.get(query_url) as response:
            self.porting_guide = await response.read()


async def collect_changelogs(
    collectors: list[CollectionChangelogCollector],
    core_collector: AnsibleCoreChangelogCollector,
    collection_cache: str | None,
    galaxy_context: GalaxyContext | None = None,
):
    lib_ctx = app_context.lib_ctx.get()
    with tempfile.TemporaryDirectory() as tmp_dir:
        async with aiohttp.ClientSession() as aio_session:
            if galaxy_context is None:
                galaxy_context = await GalaxyContext.create(aio_session)
            async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
                downloader = CollectionDownloader(
                    aio_session,
                    tmp_dir,
                    context=galaxy_context,
                    collection_cache=collection_cache,
                )

                async def core_downloader(version):
                    return await get_ansible_core(aio_session, version, tmp_dir)

                requestors = [
                    await pool.spawn(collector.download(downloader))
                    for collector in collectors
                ]
                requestors.append(
                    await pool.spawn(core_collector.download_changelog(core_downloader))
                )
                requestors.append(
                    await pool.spawn(core_collector.download_porting_guide(aio_session))
                )
                await asyncio.gather(*requestors)


class ChangelogEntry:
    version: PypiVer
    version_str: str
    is_ancestor: bool

    prev_version: PypiVer | None
    core_versions: dict[PypiVer, str]
    versions_per_collection: dict[str, dict[PypiVer, str]]

    core_collector: AnsibleCoreChangelogCollector
    ansible_changelog: ChangelogData
    collectors: list[CollectionChangelogCollector]

    ansible_core_version: str
    prev_ansible_core_version: str | None

    removed_collections: list[tuple[CollectionChangelogCollector, str]]
    added_collections: list[tuple[CollectionChangelogCollector, str]]
    unchanged_collections: list[tuple[CollectionChangelogCollector, str]]
    changed_collections: list[
        tuple[CollectionChangelogCollector, str, str | None, bool]
    ]

    def __init__(
        self,
        version: PypiVer,
        version_str: str,
        prev_version: PypiVer | None,
        ancestor_version: PypiVer | None,
        core_versions: dict[PypiVer, str],
        versions_per_collection: dict[str, dict[PypiVer, str]],
        core_collector: AnsibleCoreChangelogCollector,
        ansible_changelog: ChangelogData,
        collectors: list[CollectionChangelogCollector],
    ):
        self.version = version
        self.version_str = version_str
        self.is_ancestor = (
            False if ancestor_version is None else ancestor_version == version
        )
        self.prev_version = prev_version = prev_version or ancestor_version
        self.core_versions = core_versions
        self.versions_per_collection = versions_per_collection
        self.core_collector = core_collector
        self.ansible_changelog = ansible_changelog
        self.collectors = collectors

        self.ansible_core_version = core_versions[version]
        self.prev_ansible_core_version = (
            core_versions.get(prev_version) if prev_version else None
        )

        self.removed_collections = []
        self.added_collections = []
        self.unchanged_collections = []
        self.changed_collections = []
        for collector in collectors:
            if version not in versions_per_collection[collector.collection]:
                if (
                    prev_version
                    and prev_version in versions_per_collection[collector.collection]
                ):
                    self.removed_collections.append(
                        (
                            collector,
                            versions_per_collection[collector.collection][prev_version],
                        )
                    )

                continue

            collection_version: str = versions_per_collection[collector.collection][
                version
            ]

            prev_collection_version: str | None = (
                versions_per_collection[collector.collection].get(prev_version)
                if prev_version
                else None
            )
            added = False
            if prev_version:
                if not prev_collection_version:
                    self.added_collections.append((collector, collection_version))
                    added = True
                elif prev_collection_version == collection_version:
                    self.unchanged_collections.append((collector, collection_version))
                    continue

            self.changed_collections.append(
                (collector, collection_version, prev_collection_version, added)
            )


class Changelog:
    ansible_version: PypiVer
    ansible_ancestor_version: PypiVer | None
    entries: list[ChangelogEntry]
    core_collector: AnsibleCoreChangelogCollector
    ansible_changelog: ChangelogData
    collection_collectors: list[CollectionChangelogCollector]
    collection_metadata: CollectionsMetadata

    def __init__(
        self,
        ansible_version: PypiVer,
        ansible_ancestor_version: PypiVer | None,
        entries: list[ChangelogEntry],
        core_collector: AnsibleCoreChangelogCollector,
        ansible_changelog: ChangelogData,
        collection_collectors: list[CollectionChangelogCollector],
        collection_metadata: CollectionsMetadata,
    ):
        self.ansible_version = ansible_version
        self.ansible_ancestor_version = ansible_ancestor_version
        self.entries = entries
        self.core_collector = core_collector
        self.ansible_changelog = ansible_changelog
        self.collection_collectors = collection_collectors
        self.collection_metadata = collection_metadata


def get_changelog(
    ansible_version: PypiVer,
    deps_dir: str | None,
    deps_data: list[DependencyFileData] | None = None,
    collection_cache: str | None = None,
    ansible_changelog: ChangelogData | None = None,
    galaxy_context: GalaxyContext | None = None,
) -> Changelog:
    dependencies: dict[str, DependencyFileData] = {}

    ansible_changelog = ansible_changelog or ChangelogData.ansible(directory=deps_dir)
    ansible_ancestor_version_str = ansible_changelog.changes.ancestor
    ansible_ancestor_version = (
        PypiVer(ansible_ancestor_version_str) if ansible_ancestor_version_str else None
    )

    collection_metadata = CollectionsMetadata(deps_dir)

    if deps_dir is not None:
        for path in glob.glob(os.path.join(deps_dir, "*.deps"), recursive=False):
            deps_file = DepsFile(path)
            deps = deps_file.parse()
            deps.deps.pop("_python", None)
            version = PypiVer(deps.ansible_version)
            if version > ansible_version:
                print(
                    f"Ignoring {path}, since {deps.ansible_version}"
                    f" is newer than {ansible_version}"
                )
                continue
            dependencies[deps.ansible_version] = deps
    if deps_data:
        for deps in deps_data:
            dependencies[deps.ansible_version] = deps

    core_versions: dict[PypiVer, str] = {}
    versions: dict[str, tuple[PypiVer, DependencyFileData]] = {}
    versions_per_collection: dict[str, dict[PypiVer, str]] = defaultdict(dict)
    for deps in dependencies.values():
        version = PypiVer(deps.ansible_version)
        versions[deps.ansible_version] = (version, deps)
        core_versions[version] = deps.ansible_core_version
        for collection_name, collection_version in deps.deps.items():
            versions_per_collection[collection_name][version] = collection_version

    core_collector = AnsibleCoreChangelogCollector(core_versions.values())
    collectors = [
        CollectionChangelogCollector(
            collection, versions_per_collection[collection].values()
        )
        for collection in sorted(versions_per_collection.keys())
    ]
    asyncio.run(
        collect_changelogs(collectors, core_collector, collection_cache, galaxy_context)
    )

    changelog = []

    sorted_versions = collect_versions(versions, ansible_changelog.config)
    for index, (version_str, dummy) in enumerate(sorted_versions):
        version, deps = versions[version_str]
        prev_version = None
        if index + 1 < len(sorted_versions):
            prev_version = versions[sorted_versions[index + 1][0]][0]

        changelog.append(
            ChangelogEntry(
                version,
                version_str,
                prev_version,
                ansible_ancestor_version,
                core_versions,
                versions_per_collection,
                core_collector,
                ansible_changelog,
                collectors,
            )
        )

    return Changelog(
        ansible_version,
        ansible_ancestor_version,
        changelog,
        core_collector,
        ansible_changelog,
        collectors,
        collection_metadata,
    )



# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020

"""
Minimal setup.py for compatibility with older tools.
Configuration is in setup.cfg.
Users are recommended to use tools like pip and build and the PEP 517 interface
instead of setuptools' deprecated setup.py interface.
"""

from setuptools import setup

setup()



# Author: Felix Fontein <felix@fontein.de>
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020
"""Helper to use pkgutil.get_data without having to check the return value."""

from __future__ import annotations

import pkgutil


def get_antsibull_data(filename: str) -> bytes:
    """
    Retrieve data from the antsibull.data package as bytes.

    The filename can be a relative path separated with '/' to access subdirectories.
    See https://docs.python.org/3/library/pkgutil.html#pkgutil.get_data for details.
    """
    data = pkgutil.get_data("antsibull.data", filename)
    if data is None:
        raise RuntimeError(f"Cannot find {filename} in the antsibull.data package")
    return data



# SPDX-License-Identifier: Python-2.0.1
# Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007,
# 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019,
# 2020, 2021, 2022, 2023 Python Software Foundation; All Rights Reserved

"""
Modifed code derived from Cpython's `shutil` module to suit our own needs
"""

from __future__ import annotations

import os
import typing as t
from pathlib import Path
from shutil import Error, copy2, copystat

if t.TYPE_CHECKING:
    from _typeshed import StrPath


def copytree_and_symlinks(
    src: StrPath | os.DirEntry,
    dest: StrPath | os.DirEntry,
    *,
    ignore_dangling_symlinks: bool = True,
    dirs_exist_ok: bool = False,
    _origsrc: Path | None = None,
) -> None:
    """
    Modified version of `shutil.copytree()` that handles symlinks more inteligently.

    * Relative symlink to files within the `src` directory -> symlink is preserved
    * Relative symlink to files outside the `src` directory -> files are copied
    * `ignore_dangling_symlinks` is the default.

    Windows code branches are removed. Antsibull does not support Windows.
    """
    # Ensure that the generator is closed
    with os.scandir(src) as itr:  # type: ignore[type-var]
        entries = list(itr)
    _copytree(
        entries,
        src,
        dest,
        ignore_dangling_symlinks=ignore_dangling_symlinks,
        dirs_exist_ok=dirs_exist_ok,
        _origsrc=_origsrc,
    )


def _should_symlink(src: Path, target: Path, origsrc: Path) -> bool:
    """
    Check if `src`'s symlink target is a path outside `src` that cannot be
    symlinked

    Args:
        src:
            Source directory
        target:
            Symlink target
    """
    if target.is_absolute():
        return True
    try:
        srcr = src.resolve()
        targetr = (srcr / target).resolve()
        if not targetr.is_dir():
            targetr = targetr.parent
        is_relative = origsrc.is_relative_to(targetr)
        return origsrc == targetr or not is_relative
    except Exception:  # pragma: no cover pylint: disable=broad-exception-caught
        return False


def _copytree(  # noqa: C901
    entries: list[os.DirEntry],
    src: StrPath,
    dst: StrPath,
    ignore_dangling_symlinks: bool,
    dirs_exist_ok: bool = False,
    *,
    _origsrc: Path | None = None,
) -> None:
    copy_function = copy2
    _origsrc = _origsrc if _origsrc else Path(src).resolve()

    os.makedirs(dst, exist_ok=dirs_exist_ok)
    errors = []

    for srcentry in entries:
        srcname = os.path.join(src, srcentry.name)
        dstname = os.path.join(dst, srcentry.name)
        srcobj = srcentry
        try:
            is_symlink = srcentry.is_symlink()
            if is_symlink:
                linkto = os.readlink(srcname)
                if _should_symlink(Path(src), Path(linkto), _origsrc):
                    # We can't just leave it to `copy_function` because legacy
                    # code with a custom `copy_function` may rely on copytree
                    # doing the right thing.
                    os.symlink(linkto, dstname)
                    copystat(srcobj, dstname, follow_symlinks=False)
                else:
                    # ignore dangling symlink if the flag is on
                    if (
                        not os.path.exists(os.path.join(src, linkto))
                        and ignore_dangling_symlinks
                    ):
                        continue
                    if srcentry.is_dir():
                        copytree_and_symlinks(
                            srcobj,
                            dstname,
                            ignore_dangling_symlinks=ignore_dangling_symlinks,
                            dirs_exist_ok=dirs_exist_ok,
                            _origsrc=_origsrc,
                        )
                    else:
                        copy_function(srcobj, dstname)
            elif srcentry.is_dir():
                copytree_and_symlinks(
                    srcobj,
                    dstname,
                    ignore_dangling_symlinks=ignore_dangling_symlinks,
                    dirs_exist_ok=dirs_exist_ok,
                    _origsrc=_origsrc,
                )
            else:
                # Will raise a SpecialFileError for unsupported file types
                copy_function(srcobj, dstname)
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except Error as err:
            errors.extend(err.args[0])
        except OSError as why:
            errors.append((srcname, dstname, str(why)))
    copystat(src, dst)
    if errors:
        raise Error(errors)


__all__ = ("copytree_and_symlinks",)



# Copyright (C) 2021 David Moreau-Simard <dmsimard@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import yaml
import ansible
from ansible.plugins.action import ActionBase

DEFAULT_RUNTIME = "%s/config/ansible_builtin_runtime.yml" % os.path.dirname(ansible.__file__)

DOCUMENTATION = """
---
module: parse_ansible_runtime
short_description: Parses an ansible_builtin_runtime.yml file for testing purposes.
version_added: "2.12"
author: "David Moreau-Simard <dmsimard@redhat.com>"
description:
    - Parses an ansible_builtin_runtime.yml file for testing purposes.
options:
    runtime_file:
        description:
            - Path to an ansible_builtin_runtime.yml file
            - If not set, the module will default to the one for the currently-running Ansible process.
        required: false
"""

EXAMPLES = """
- name: Parse the runtime currently used by Ansible
  parse_ansible_runtime:
  register: _parsed_runtime

- name: Parse a specific runtime configuration file
  parse_ansible_runtime:
    runtime_file: /opt/git/ansible/lib/ansible/config/ansible_builtin_runtime.yml
  register: _parsed_runtime

- name: List parsed collections
  debug:
    msg: "{{ item }}"
  loop: "{{ _parsed_runtime.collections }}"

- name: List parsed modules (redirection targets)
  debug:
    msg: "{{ item }}"
  loop: "{{ _parsed_runtime.modules }}"
"""

RETURN = """
collections:
    description: List of collection names found in module redirections
    returned: on success
    type: list
    sample: ['community.crypto', 'community.general']
modules:
    description: List of module names found in module redirections
    returned: on success
    type: list
    sample: ['community.general.ovirt', 'community.general.proxmox']
"""


class ActionModule(ActionBase):
    """ Parses an Ansible runtime file from the Ansible controller """

    TRANSFERS_FILES = False

    def __init__(self, *args, **kwargs):
        super(ActionModule, self).__init__(*args, **kwargs)

    def run(self, tmp=None, task_vars=None):
        if task_vars is None:
            task_vars = dict()

        result = super(ActionModule, self).run(tmp, task_vars)
        runtime_file = self._task.args.get("runtime_file", DEFAULT_RUNTIME)
        runtime = yaml.safe_load(open(runtime_file, "r"))

        # TODO: Add other verifiable plugin types
        collections = []
        modules = []
        for module, target in runtime["plugin_routing"]["modules"].items():
            if "redirect" in target:
                collections.append(".".join(target["redirect"].split('.')[:-1]))
                modules.append(target["redirect"])

        result["collections"] = sorted(list(set(collections)))
        result["modules"] = sorted(list(set(modules)))
        result["changed"] = False
        result["msg"] = "Parsed %s successfully" % runtime_file

        return result



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import contextlib
from collections.abc import AsyncIterator, Iterator
from pathlib import Path
from unittest.mock import MagicMock

import pytest

from antsibull._vendor.shutil import _should_symlink
from antsibull.utils.paths import atemp_or_dir, copytree_and_symlinks, temp_or_dir


@pytest.mark.asyncio
async def test_atemp_or_dir_new(monkeypatch: pytest.MonkeyPatch) -> None:
    @contextlib.asynccontextmanager
    async def mock_cm() -> AsyncIterator[str]:
        yield "abc"

    monkeypatch.setattr("aiofiles.tempfile.TemporaryDirectory", mock_cm)
    async with atemp_or_dir() as temp:
        assert temp == Path("abc")


@pytest.mark.asyncio
async def test_atemp_or_dir(tmp_path: Path) -> None:
    async with atemp_or_dir(tmp_path) as temp:
        assert temp == tmp_path


def test_temp_or_dir_new(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(
        "tempfile.TemporaryDirectory", lambda: contextlib.nullcontext("abc")
    )
    with temp_or_dir() as temp:
        assert temp == Path("abc")


def test_temp_or_dir(tmp_path: Path) -> None:
    with temp_or_dir(tmp_path) as temp:
        assert temp == tmp_path


def test_temp_not_dir(tmp_path: Path) -> None:
    abc = str(tmp_path / "abc")
    with pytest.raises(ValueError, match=f"{abc} is not a directory!"):
        temp_or_dir(abc).__enter__()


def test_should_symlink_relative(tmp_path: Path) -> None:
    here = tmp_path / "src" / "123"
    here.mkdir(parents=True)
    (here.parent / "xyz").touch()
    (here / "test_dir").mkdir()
    assert not _should_symlink(here, Path("test_dir/../../"), here)
    assert _should_symlink(here, Path("test_dir/../../123"), here)


def test_should_symlink_absolute() -> None:
    # This should never be accessed if the symlink is absolute
    fake_path = MagicMock(side_effect=Exception)
    #
    assert _should_symlink(Path("src/abc"), Path("/bin/bash"), fake_path)


def test_should_symlink_simple(tmp_path: Path) -> None:
    here = tmp_path / "src" / "123"
    here.mkdir(parents=True)
    (here / "xyz").touch()
    assert _should_symlink(here, Path("xyz"), here)

    (here2 := here / "another_directory").mkdir()
    (here2 / "456").touch()
    assert _should_symlink(here, Path("another_directory/456"), here)


def _pathiter(directory: Path, /) -> Iterator[Path]:
    for path in directory.iterdir():
        if path.is_dir() and not path.is_symlink():
            yield path
            yield from _pathiter(path)
        else:
            yield path


def test_copytree_functional(
    tmp_path: Path, test_data_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    old = test_data_path / "copytree"
    new = tmp_path / "newtree"
    copytree_and_symlinks(old, new)

    new_paths = list(_pathiter(new))
    assert len(new_paths) == 8
    monkeypatch.chdir(new)
    assert Path("a2").readlink() == Path("a")
    assert Path("recursive").readlink() == Path()
    assert Path("README.md").is_file() and not Path("README.md").is_symlink()



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)


"""
Utilities to run sanity tests accross collections
"""

from __future__ import annotations

import datetime
import json
import shlex
import shutil
import sys
from collections.abc import Collection, Iterable, Iterator, Sequence
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from subprocess import CompletedProcess
from typing import TYPE_CHECKING, Any, TypedDict

from antsibull_core import app_context
from antsibull_core.logging import log
from antsibull_core.subprocess_util import log_run
from antsibull_core.yaml import store_yaml_file
from packaging.version import Version

from antsibull.constants import SANITY_TESTS_BANNED_IGNORES, SANITY_TESTS_DEFAULT
from antsibull.types import CollectionName, add_dataclass_yaml_type

if TYPE_CHECKING:
    from _typeshed import StrPath

mlog = log.fields(mod=__name__)


class SanityOutput(TypedDict):
    """
    Mapping of `ansible-test sanity` output and other related data
    """

    cmd: list[str]
    returncode: int
    stdout: str
    stderr: str
    runtime: float
    test_json: dict[str, Any]
    ignore_entries: list[IgnoreEntry]
    banned_ignore_entries: list[IgnoreEntry]
    ignores_file: str | None


class CollectionOutput(TypedDict):
    """
    Collection entry
    """

    failed: bool
    sanity: SanityOutput


class EnvDetails(TypedDict):
    """
    Details about the ansible-test environment
    """

    ansible_test_version: str
    sanity_tests: list[str]


class Output(TypedDict):
    """
    Parent mapping of the sanity test data file
    """

    collections: dict[CollectionName, CollectionOutput]
    total_runtime: float
    env_details: EnvDetails


@dataclass(frozen=True)
class IgnoreEntry:
    """
    Represents an entry in an `ansible-test sanity` ignores file
    """

    file: str
    test: str
    remainder: str

    def as_str(self) -> str:
        return f"{self.file} {self.test}{self.remainder}"


add_dataclass_yaml_type(IgnoreEntry)


def parse_ignores_file(file: StrPath) -> Iterator[IgnoreEntry]:
    """
    Parse a sanity test ignore file
    """
    flog = mlog.fields(func="parse_sanity_test_file", file=file)
    with open(file, encoding="utf-8") as fp:
        for line in fp:
            line = line.rstrip("\n")
            parsed = line.split(" ", 2)
            if len(parsed) == 2:
                parsed.append("")
            if len(parsed) != 3:
                flog.error("Failed to parse line: {0}", line)
                continue
            yield IgnoreEntry(file=parsed[0], test=parsed[1], remainder=parsed[2])


def get_ignores_file(directory: Path, version: str) -> Path | None:
    """
    Determine the path to a sanity test ignore file for a certain ansible-test `version`

    Args:
        directory:
            Collection's directory
        version:
            Version of ansible-test

    Returns:
        Path to a sanity test ignore file if one exists or `None`
    """
    expected = directory / "tests/sanity" / f"ignore-{version}.txt"
    return expected if expected.is_file() else None


def filter_invalid_ignores(
    entries: Iterable[IgnoreEntry],
    matches: Collection[str] = SANITY_TESTS_BANNED_IGNORES,
) -> Iterator[IgnoreEntry]:
    """
    Given an Iterable of `IgnoreEntry`s, filter out entries that match `matches`

    Args:
        matches:
            Iterable of `IgnoreEntry`s
        invalid:
            `Collection` of sanity test ignore names
    Yields:
        `IgnoreEntry`s that match `matches`
    """
    for entry in entries:
        if entry.test in matches:
            yield entry


def is_git_repo(base_directory: Path, directory: Path) -> bool:
    """
    Check if a collection directory is its own git directory

    Args:
        base_directory:
            Base directory that contains the collection directory
        directory:
            Collection directory

    Returns:
        Whether or not the collection directory is its own git directory
    """
    repo = log_run(["git", "rev-parse", "--git-dir"], check=False, cwd=directory)
    if repo.returncode == 128:
        return False
    git_directory = Path(repo.stdout.strip())
    if not git_directory.is_absolute():
        git_directory = directory.resolve() / git_directory
    # If the collection base directory is a subdirectory of the git repository,
    # we need to create an empty one for the collection.
    return not base_directory.resolve().is_relative_to(git_directory.parent)


def create_stub_git_repo(directory: StrPath) -> None:
    """
    Create a git repository and initial commit with all files in a directory
    """
    git_run = partial(log_run, cwd=directory)
    git_run(["git", "init", "."])
    git_run(["git", "add", "."])
    git_run(["git", "commit", "--no-gpg-sign", "-m", "init"])


def remove_test_output_files(directory: Path) -> None:
    """
    Clean ansible-test output directory
    """
    if (output := directory / "tests" / "output").exists():
        shutil.rmtree(output)


def get_errors(directory: Path) -> dict[str, Any]:
    """
    Get `ansible-test sanity` error JSON output
    """
    files = directory.glob("tests/output/bot/ansible-test-sanity-*.json")
    data: dict[str, Any] = {}
    for file in files:
        with file.open() as fp:
            data[file.name] = json.load(fp)
    return data


def get_ansible_test_version(ansible_test_bin: Sequence[StrPath]) -> str:
    """
    Get the ansible-test version from `ansible_test_bin`
    """
    args = [*ansible_test_bin, "--version"]
    stdout = log_run(args).stdout.strip()
    version = stdout.split(" ")[-1]
    # Ensure version is parsable
    Version(version)
    return version


def run_sanity_tests(
    collection: CollectionName,
    directory: Path,
    tests: Sequence[str],
    quiet: bool,
    ansible_test_bin: Sequence[StrPath] = ("ansible-test",),
) -> tuple[CompletedProcess, float]:
    """
    Run `ansible-test sanity` and log
    """
    test_args = [f"--test={test}" for test in tests]
    args = [
        *ansible_test_bin,
        "sanity",
        *test_args,
        "--docker",
        "--lint",
    ]
    print(f"* Running: {args} for {collection}", file=sys.stderr)

    now = datetime.datetime.now()
    cmd = log_run(
        args,
        # Always show stderr
        stderr_loglevel="debug" if quiet else "error",
        cwd=directory,
        check=False,
    )
    after = datetime.datetime.now()
    total = after - now
    runtime = total.seconds / 60
    failed = "FAILURE: " if cmd.returncode else ""
    print(f"* {failed}Finished in {runtime:.2f} minutes\n", file=sys.stderr)
    return cmd, runtime


def _get_ignores_info(
    directory: Path, ansible_test_version: str
) -> tuple[list[IgnoreEntry], list[IgnoreEntry], str | None]:
    v = Version(ansible_test_version)
    ignore_file = get_ignores_file(directory, f"{v.major}.{v.minor}")
    ignores: list[IgnoreEntry] = []
    banned_ignores: list[IgnoreEntry] = []
    file_base: str | None = None
    if ignore_file:
        ignores = list(parse_ignores_file(ignore_file))
        banned_ignores = list(filter_invalid_ignores(ignores))
        file_base = str(ignore_file.relative_to(directory))
    return ignores, banned_ignores, file_base


def handle_collection(
    directory: Path,
    tests: Sequence[str],
    clean: bool,
    quiet: bool,
    env_details: EnvDetails,
    ansible_test_bin: Sequence[StrPath] = ("ansible-test",),
) -> tuple[CollectionName, CollectionOutput]:
    """
    Run the tests for a collection
    """
    collection = CollectionName.from_galaxy_yml(directory / "galaxy.yml")
    if clean:
        remove_test_output_files(directory)
    if not is_git_repo(directory.parent, directory):
        create_stub_git_repo(directory)
    cmd, runtime = run_sanity_tests(
        collection, directory, tests, quiet, ansible_test_bin
    )
    errors = get_errors(directory)
    ignores, banned_ignores, ignores_file = _get_ignores_info(
        directory, env_details["ansible_test_version"]
    )
    sanity_output = SanityOutput(
        cmd=cmd.args,
        returncode=cmd.returncode,
        stdout=cmd.stdout,
        stderr=cmd.stderr,
        runtime=runtime,
        test_json=errors,
        ignore_entries=ignores,
        banned_ignore_entries=banned_ignores,
        ignores_file=ignores_file,
    )
    return (
        collection,
        # Namespace the data under "sanity" for futureproofing
        {
            "sanity": sanity_output,
            "failed": bool(cmd.returncode) or bool(banned_ignores),
        },
    )


def get_percentage(collections: dict[CollectionName, CollectionOutput]) -> str:
    """
    Determine success/failure percentage
    """
    colls_succeeded = [cdata for cdata in collections.values() if not cdata["failed"]]
    perc = f"{len(colls_succeeded)/len(collections)*100:.2f}%"
    frac = f"{len(colls_succeeded)} / {len(collections)}"
    return f"{perc} ({frac}) succeeded"


def sanity_tests_command() -> int:
    """
    Run sanity tests across multiple collections and store results
    """
    app_ctx = app_context.app_ctx.get()
    collections: list[Path] = app_ctx.extra["collection_paths"]
    tests: list[str] = app_ctx.extra["tests"] or list(SANITY_TESTS_DEFAULT)
    clean: bool = app_ctx.extra["clean"]
    quiet: bool = app_ctx.extra["quiet"]
    error_output: Path = app_ctx.extra["error_output"]
    # Allow spaces for something like "python -m ansible test"
    ansible_test_bin: list[str] = shlex.split(app_ctx.extra["ansible_test_bin"])

    env_details: EnvDetails = {
        "ansible_test_version": get_ansible_test_version(ansible_test_bin),
        "sanity_tests": tests,
    }

    collections_errors: dict[CollectionName, CollectionOutput] = dict(
        handle_collection(
            collection, tests, clean, quiet, env_details, ansible_test_bin
        )
        for collection in collections
    )
    total_runtime = sum(
        cerrors["sanity"]["runtime"] for cerrors in collections_errors.values()
    )
    data: Output = {
        "collections": collections_errors,
        "total_runtime": total_runtime,
        "env_details": env_details,
    }

    print(get_percentage(collections_errors), f"in {total_runtime:.2f} minutes!")
    store_yaml_file(error_output, data)
    return 0 if any(c["failed"] for c in collections_errors.values()) else 1



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Commands to clone and verify collections from source
"""

from __future__ import annotations

import asyncio
import json
from collections.abc import Collection, Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Any

import asyncio_pool  # type: ignore[import]
from antsibull_core import app_context
from antsibull_core.subprocess_util import async_log_run
from antsibull_core.yaml import load_yaml_file, store_yaml_file

from antsibull.build_ansible_commands import download_collections
from antsibull.tagging import CollectionTagData
from antsibull.types import CollectionName, make_collection_mapping
from antsibull.utils.galaxy import create_galaxy_context
from antsibull.utils.paths import atemp_or_dir

from ._utils import filter_tag_data, tag_data_as_version_mapping
from .clone import NormalizedCheckout, clone_collection, normalize_clone
from .verify import FileError, verify_files

if TYPE_CHECKING:
    from _typeshed import StrPath


async def _extract_files_data(collection: StrPath) -> list[dict[str, Any]]:
    data = (await async_log_run(["tar", "-Oxf", collection, "FILES.json"])).stdout
    return json.loads(data)["files"]


async def _clone_collections(
    collections: dict[CollectionName, CollectionTagData], download_dir: Path
) -> dict[CollectionName, Path]:
    for namespace in {col.namespace for col in collections}:
        (download_dir / namespace).mkdir()
    lib_ctx = app_context.lib_ctx.get()
    async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
        requestors = {
            collection: await pool.spawn(
                clone_collection(collection, data, download_dir)
            )
            for collection, data in collections.items()
        }
        values = await asyncio.gather(*requestors.values())
    final: dict[CollectionName, Path] = dict(zip(collections.keys(), values))
    return final


async def _handle_collection(
    collection: CollectionName,
    tag_data: CollectionTagData,
    artifact: Path,
    checkout: Path,
    ignore_errors: Collection[FileError] | None,
) -> NormalizedCheckout:
    norm = await normalize_clone(collection, checkout, tag_data)
    norm.errors.extend(
        [
            output
            async for output in verify_files(
                collection,
                checkout / norm.collection_subdir,
                (await _extract_files_data(artifact)),
                ignore_errors,
            )
        ]
    )
    return norm


async def _handle_collections(
    tag_data: dict[CollectionName, CollectionTagData],
    artifacts: dict[CollectionName, str],
    checkouts: dict[CollectionName, Path],
    ignore_errors: Collection[FileError] | None,
) -> list[NormalizedCheckout]:
    lib_ctx = app_context.lib_ctx.get()
    async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
        tasks = [
            await pool.spawn(
                _handle_collection(
                    collection,
                    tag_data[collection],
                    Path(artifacts[collection]),
                    checkouts[collection],
                    ignore_errors,
                )
            )
            for collection in tag_data
        ]
        return list(await asyncio.gather(*tasks))


async def _symlink_tree(
    checkouts: Iterable[NormalizedCheckout], tree_path: Path, allow_symlink: bool = True
):
    lib_ctx = app_context.lib_ctx.get()
    for namespace in {col.collection.namespace for col in checkouts}:
        (tree_path / namespace).mkdir()
    async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
        tasks = [
            await pool.spawn(
                collection.add_to_collection_tree(tree_path, allow_symlink)
            )
            for collection in checkouts
        ]
        await asyncio.gather(*tasks)


#####


def verify_upstream_command() -> int:
    """
    Ensure that files in a collection's git repository (roughly) match the
    contents of its upstream repository
    """
    return asyncio.run(_verify_upstream_command())


async def _verify_upstream_command() -> int:
    app_ctx = app_context.app_ctx.get()
    lib_ctx = app_context.lib_ctx.get()
    tags_file: Path = app_ctx.extra["tags_file"]
    globs: list[str] | None = app_ctx.extra["globs"]
    download_dir: Path | None = app_ctx.extra["download_dir"]
    checkouts_dir_: Path | None = app_ctx.extra["checkouts_dir"]
    tree_dir: Path | None = app_ctx.extra["tree_dir"]
    ignores: Collection[FileError] = app_ctx.extra["ignores"]
    error_output: Path = app_ctx.extra["error_output"]

    tags_data: dict[CollectionName, CollectionTagData] = make_collection_mapping(
        load_yaml_file(tags_file)
    )
    tags_data = filter_tag_data(tags_data, globs)
    versions = tag_data_as_version_mapping(tags_data)
    async with atemp_or_dir(download_dir) as download_dir:
        artifacts_dir = download_dir / "collection_artifacts"
        artifacts_dir.mkdir(parents=True)
        checkouts_dir = Path(
            (checkouts_dir_ or (download_dir / "checkouts")),
            "ansible_collections",
        )
        checkouts_dir.mkdir(parents=True)
        if tree_dir:
            tree_dir /= "ansible_collections"
            tree_dir.mkdir(parents=True)

        artifacts = make_collection_mapping(
            await download_collections(
                versions,  # type: ignore[arg-type]
                str(artifacts_dir),
                await create_galaxy_context(),
                lib_ctx.collection_cache,
            )
        )
        checkouts = await _clone_collections(tags_data, checkouts_dir)

        normed_collections = await _handle_collections(
            tags_data, artifacts, checkouts, ignores
        )
        error_blob = {
            "collections": {
                coll.collection: coll.errors
                for coll in normed_collections
                if coll.errors
            }
        }

        store_yaml_file(error_output, error_blob)

        if tree_dir:
            # Create symlinks to collections in checkouts_dir instead of
            # copying if checkouts_dir is explicitly passed by the user as
            # opposed to being a temporary download directory.
            allow_symlink = bool(checkouts_dir_)
            #
            await _symlink_tree(normed_collections, tree_dir, allow_symlink)
    return bool(error_blob)


__all__ = ("verify_upstream_command",)



# Author: Felix Fontein <felix@fontein.de>
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020

"""
Classes to encapsulate collection metadata from collection-meta.yaml
"""

from __future__ import annotations

import os
import typing as t
from collections.abc import Mapping

from antsibull_core.yaml import load_yaml_file


class CollectionMetadata:
    """
    Stores metadata about one collection.
    """

    changelog_url: str | None
    collection_directory: str | None
    repository: str | None
    tag_version_regex: str | None

    def __init__(self, source: Mapping[str, t.Any] | None = None):
        if source is None:
            source = {}
        self.changelog_url = source.get("changelog-url")
        self.collection_directory = source.get("collection-directory")
        self.repository = source.get("repository")
        self.tag_version_regex = source.get("tag_version_regex")


class CollectionsMetadata:
    """
    Stores metadata about a set of collections.
    """

    data: dict[str, CollectionMetadata]

    def __init__(self, deps_dir: str | None):
        self.data = {}
        if deps_dir is not None:
            collection_meta_path = os.path.join(deps_dir, "collection-meta.yaml")
            if os.path.exists(collection_meta_path):
                data = load_yaml_file(collection_meta_path)
                if data and "collections" in data:
                    for collection_name, collection_data in data["collections"].items():
                        self.data[collection_name] = CollectionMetadata(collection_data)

    def get_meta(self, collection_name: str) -> CollectionMetadata:
        result = self.data.get(collection_name)
        if result is None:
            result = CollectionMetadata()
            self.data[collection_name] = result
        return result



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Verify collection source trees
"""

from __future__ import annotations

from collections.abc import AsyncIterator, Collection
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypedDict

import aiofiles.ospath
from antsibull_core.utils.hashing import verify_hash

from antsibull.types import add_yaml_type

if TYPE_CHECKING:
    from typing_extensions import NotRequired


class FileError(str, Enum):
    """
    Collections source verification errors
    """

    # A file exists in the repo but its hash is invalid
    WRONG_HASH = "WRONG_HASH"
    # A file does not exist in the repo
    MISSING_FILE = "MISSING_FILE"
    # A directory in the artifact is a regular file in the repo
    NOT_A_DIRECTORY = "NOT_A_DIRECTORY"
    # The collection namespace/name in galaxy.yml does not match
    WRONG_GALAXY_YML_NAME = "WRONG_GALAXY_YML_NAME"

    def __str__(self) -> str:
        return self.name


LENIENT_FILE_ERROR_IGNORES = frozenset({FileError.MISSING_FILE})


add_yaml_type(FileError)


class FileErrorOutput(TypedDict):
    """
    Mapping of a FileError and the file name
    """

    file: str
    error: FileError
    message: NotRequired[str]


async def verify_files(
    collection: str,  # pylint: disable=unused-argument
    collection_dir: Path,
    files_data: list[dict[str, Any]],
    ignore_errors: Collection[FileError] | None = None,
) -> AsyncIterator[FileErrorOutput]:
    """
    Ensure that files in a collection git repository checkout match the
    artifact's FILES.json metadata

    Args:
        collection:
            Name of the collection being tested
        checkout_dir:
            Directory containing a collection checkout
        files_data:
            The `files` list from a collection artifact's FILES.json metadata
        ignore_errors:
            Collection of `FileError`s to ignore to ignore.
            Defaults to `{MISSING_FILE}`.

    Yields:
        `FileErrorOutput` dicts with error data
    """
    for file in files_data:
        path = collection_dir / file["name"]
        is_dir = file["ftype"] == "dir"
        error: FileError | None = None
        if not await aiofiles.ospath.exists(path):
            error = FileError.MISSING_FILE
        elif is_dir and not await aiofiles.ospath.isdir(path):
            error = FileError.NOT_A_DIRECTORY
        elif not is_dir and not await verify_hash(
            path, file["chksum_sha256"], "sha256"
        ):
            error = FileError.WRONG_HASH
        if error is not None and error not in (ignore_errors or set()):
            yield FileErrorOutput({"file": file["name"], "error": error})



# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+
# (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import filecmp
import subprocess
import sys
from pathlib import Path

import pytest

from antsibull.cli.antsibull_build import run

ANNOUNCEMENT_TESTS = [
    ("7.0.0b1", "announce-7.0.0b1"),
    ("7.0.0", "announce-7.0.0"),
    ("7.4.0", "announce-7.4.0"),
]

COLOR_BOLD = "\x1B[0;97m\x1B[1m"
COLOR_NORMAL = "\x1B[0m"


@pytest.mark.parametrize("version, directory", ANNOUNCEMENT_TESTS)
def test_announcements_command(
    test_data_path: Path,
    tmp_path_factory: pytest.TempPathFactory,
    version: str,
    directory: str,
) -> None:
    result_path = tmp_path_factory.mktemp(f"announcements-{version}")
    ret = run(
        [
            "antsibull-build",
            "announcements",
            f"--data-dir={test_data_path}",
            f"-O{result_path}",
            version,
        ]
    )
    assert ret == 0

    expected = test_data_path / directory

    expected_files = set()
    failures = False
    for file in expected.iterdir():
        resulting_file = result_path / file.name
        if not resulting_file.exists():
            failures = True
            print(
                f"❌  {COLOR_BOLD}{file.name}{COLOR_NORMAL}: {resulting_file} does not exist",
                file=sys.stderr,
            )
        elif not filecmp.cmp(file, resulting_file):
            failures = True
            print(
                f"❌  {COLOR_BOLD}{file.name}{COLOR_NORMAL}: {resulting_file} differs from {file}",
                file=sys.stderr,
            )
            subprocess.call(
                ["diff", "--unified", "--color=always", str(file), str(resulting_file)]
            )
        expected_files.add(file.name)
    for file in result_path.iterdir():
        if file.name not in expected_files:
            failures = True
            print(
                f"❌  {COLOR_BOLD}{file.name}{COLOR_NORMAL} should not have been generated",
                file=sys.stderr,
            )

    if failures:
        print(
            "⚠️  If you updated the announcement messages and the tests are now failing, "
            f"you can run {COLOR_BOLD}cp {result_path}/* {expected}/{COLOR_NORMAL} to copy the generated files "
            "to the test fixtures. Then you can use `git diff` to see the differences.",
            file=sys.stderr,
        )

    assert not failures



#!/usr/bin/python -tt
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020

import os
import sys
from setuptools import setup


__version__ = '7.5.0'
__author__ = 'Ansible, Inc.'


with open('README.rst', 'r') as f:
    long_desc = f.read()

setup(
    name='ansible',
    version=__version__,
    description='Radically simple IT automation',
    long_description=long_desc,
    author=__author__,
    author_email='info@ansible.com',
    url='https://ansible.com/',
    project_urls={
        'Bug Tracker': 'https://github.com/ansible/ansible/issues',
        'Code of Conduct': 'https://docs.ansible.com/ansible/latest/community/code_of_conduct.html',
        'Documentation': 'https://docs.ansible.com/ansible/',
        'Mailing lists': 'https://docs.ansible.com/ansible/latest/community/communication.html#mailing-list-information',
        'Source Code': 'https://github.com/ansible/ansible',
    },
    license='GPLv3+',
    python_requires='>=3.9',
    packages=['ansible_collections'],
    exclude_package_data={
        'ansible_collections': [
            'amazon/aws/.*',
            'amazon/aws/docs/*',
            'amazon/aws/tests/*',
            'ansible/netcommon/.*',
            'ansible/netcommon/docs/*',
            'ansible/netcommon/tests/*',
            'ansible/posix/.*',
            'ansible/posix/docs/*',
            'ansible/posix/tests/*',
            'ansible/utils/.*',
            'ansible/utils/docs/*',
            'ansible/utils/tests/*',
            'ansible/windows/.*',
            'ansible/windows/docs/*',
            'ansible/windows/tests/*',
            'arista/eos/.*',
            'arista/eos/docs/*',
            'arista/eos/tests/*',
            'awx/awx/tests/*',
            'azure/azcollection/.*',
            'azure/azcollection/tests/*',
            'check_point/mgmt/tests/*',
            'chocolatey/chocolatey/tests/*',
            'cisco/aci/.*',
            'cisco/aci/tests/*',
            'cisco/asa/.*',
            'cisco/asa/docs/*',
            'cisco/asa/tests/*',
            'cisco/dnac/.*',
            'cisco/dnac/docs/*',
            'cisco/dnac/tests/*',
            'cisco/intersight/.*',
            'cisco/ios/.*',
            'cisco/ios/docs/*',
            'cisco/ios/tests/*',
            'cisco/iosxr/.*',
            'cisco/iosxr/docs/*',
            'cisco/iosxr/tests/*',
            'cisco/ise/.*',
            'cisco/ise/docs/*',
            'cisco/ise/tests/*',
            'cisco/meraki/.*',
            'cisco/meraki/tests/*',
            'cisco/mso/.*',
            'cisco/mso/tests/*',
            'cisco/nso/.*',
            'cisco/nso/tests/*',
            'cisco/nxos/.*',
            'cisco/nxos/docs/*',
            'cisco/nxos/tests/*',
            'cisco/ucs/.*',
            'cloud/common/tests/*',
            'cloudscale_ch/cloud/.*',
            'cloudscale_ch/cloud/tests/*',
            'community/aws/.*',
            'community/aws/docs/*',
            'community/aws/tests/*',
            'community/azure/.*',
            'community/azure/tests/*',
            'community/ciscosmb/.*',
            'community/ciscosmb/tests/*',
            'community/crypto/.*',
            'community/crypto/docs/*',
            'community/crypto/tests/*',
            'community/digitalocean/.*',
            'community/digitalocean/tests/*',
            'community/dns/.*',
            'community/dns/docs/*',
            'community/dns/tests/*',
            'community/docker/.*',
            'community/docker/docs/*',
            'community/docker/tests/*',
            'community/fortios/.*',
            'community/fortios/tests/*',
            'community/general/.*',
            'community/general/docs/*',
            'community/general/tests/*',
            'community/google/.*',
            'community/google/tests/*',
            'community/grafana/.*',
            'community/grafana/tests/*',
            'community/hashi_vault/.*',
            'community/hashi_vault/docs/*',
            'community/hashi_vault/tests/*',
            'community/hrobot/.*',
            'community/hrobot/docs/*',
            'community/hrobot/tests/*',
            'community/libvirt/.*',
            'community/libvirt/tests/*',
            'community/mongodb/.*',
            'community/mysql/.*',
            'community/mysql/tests/*',
            'community/network/.*',
            'community/network/tests/*',
            'community/okd/.*',
            'community/okd/docs/*',
            'community/okd/tests/*',
            'community/postgresql/.*',
            'community/postgresql/docs/*',
            'community/postgresql/tests/*',
            'community/proxysql/.*',
            'community/proxysql/tests/*',
            'community/rabbitmq/.*',
            'community/rabbitmq/tests/*',
            'community/routeros/.*',
            'community/routeros/docs/*',
            'community/routeros/tests/*',
            'community/sap/.*',
            'community/sap/tests/*',
            'community/sap_libs/.*',
            'community/sap_libs/tests/*',
            'community/skydive/.*',
            'community/skydive/tests/*',
            'community/sops/.*',
            'community/sops/docs/*',
            'community/sops/tests/*',
            'community/vmware/.*',
            'community/vmware/docs/*',
            'community/vmware/tests/*',
            'community/windows/.*',
            'community/windows/docs/*',
            'community/windows/tests/*',
            'community/zabbix/.*',
            'community/zabbix/docs/*',
            'community/zabbix/tests/*',
            'containers/podman/docs/*',
            'containers/podman/tests/*',
            'cyberark/conjur/.*',
            'cyberark/conjur/tests/*',
            'cyberark/pas/.*',
            'cyberark/pas/docs/*',
            'cyberark/pas/tests/*',
            'dellemc/enterprise_sonic/.*',
            'dellemc/enterprise_sonic/tests/*',
            'dellemc/openmanage/.*',
            'dellemc/openmanage/docs/*',
            'dellemc/openmanage/tests/*',
            'dellemc/os10/.*',
            'dellemc/os10/docs/*',
            'dellemc/os10/tests/*',
            'dellemc/os6/.*',
            'dellemc/os6/docs/*',
            'dellemc/os6/tests/*',
            'dellemc/os9/.*',
            'dellemc/os9/docs/*',
            'dellemc/os9/tests/*',
            'dellemc/powerflex/.*',
            'dellemc/powerflex/docs/*',
            'dellemc/powerflex/tests/*',
            'dellemc/unity/.*',
            'dellemc/unity/docs/*',
            'dellemc/unity/tests/*',
            'f5networks/f5_modules/docs/*',
            'f5networks/f5_modules/tests/*',
            'fortinet/fortimanager/tests/*',
            'frr/frr/.*',
            'frr/frr/docs/*',
            'frr/frr/tests/*',
            'gluster/gluster/.*',
            'gluster/gluster/tests/*',
            'google/cloud/.*',
            'google/cloud/tests/*',
            'grafana/grafana/.*',
            'grafana/grafana/tests/*',
            'hetzner/hcloud/.*',
            'hetzner/hcloud/tests/*',
            'hpe/nimble/.*',
            'hpe/nimble/tests/*',
            'ibm/qradar/.*',
            'ibm/qradar/docs/*',
            'ibm/qradar/tests/*',
            'ibm/spectrum_virtualize/.*',
            'ibm/spectrum_virtualize/docs/*',
            'ibm/spectrum_virtualize/tests/*',
            'infinidat/infinibox/.*',
            'infinidat/infinibox/docs/*',
            'infinidat/infinibox/tests/*',
            'infoblox/nios_modules/.*',
            'infoblox/nios_modules/tests/*',
            'inspur/ispim/.*',
            'inspur/ispim/docs/*',
            'inspur/sm/.*',
            'inspur/sm/docs/*',
            'junipernetworks/junos/.*',
            'junipernetworks/junos/docs/*',
            'junipernetworks/junos/tests/*',
            'kubernetes/core/.*',
            'kubernetes/core/docs/*',
            'kubernetes/core/tests/*',
            'lowlydba/sqlserver/.*',
            'lowlydba/sqlserver/docs/*',
            'lowlydba/sqlserver/tests/*',
            'mellanox/onyx/.*',
            'mellanox/onyx/tests/*',
            'microsoft/ad/.*',
            'microsoft/ad/docs/*',
            'microsoft/ad/tests/*',
            'netapp/aws/.*',
            'netapp/aws/tests/*',
            'netapp/azure/.*',
            'netapp/azure/tests/*',
            'netapp/cloudmanager/.*',
            'netapp/cloudmanager/tests/*',
            'netapp/elementsw/.*',
            'netapp/elementsw/tests/*',
            'netapp/ontap/.*',
            'netapp/ontap/tests/*',
            'netapp/storagegrid/.*',
            'netapp/storagegrid/tests/*',
            'netapp/um_info/.*',
            'netapp/um_info/tests/*',
            'netapp_eseries/santricity/.*',
            'netapp_eseries/santricity/tests/*',
            'netbox/netbox/.*',
            'netbox/netbox/docs/*',
            'netbox/netbox/tests/*',
            'ngine_io/cloudstack/.*',
            'ngine_io/cloudstack/tests/*',
            'ngine_io/exoscale/.*',
            'ngine_io/exoscale/tests/*',
            'ngine_io/vultr/.*',
            'ngine_io/vultr/tests/*',
            'openstack/cloud/docs/*',
            'openvswitch/openvswitch/.*',
            'openvswitch/openvswitch/docs/*',
            'openvswitch/openvswitch/tests/*',
            'ovirt/ovirt/.*',
            'ovirt/ovirt/tests/*',
            'purestorage/flasharray/.*',
            'purestorage/flasharray/docs/*',
            'purestorage/flasharray/tests/*',
            'purestorage/flashblade/.*',
            'purestorage/flashblade/tests/*',
            'purestorage/fusion/.*',
            'purestorage/fusion/tests/*',
            'sensu/sensu_go/.*',
            'sensu/sensu_go/docs/*',
            'sensu/sensu_go/tests/*',
            'splunk/es/.*',
            'splunk/es/docs/*',
            'splunk/es/tests/*',
            't_systems_mms/icinga_director/.*',
            't_systems_mms/icinga_director/docs/*',
            'vmware/vmware_rest/.*',
            'vmware/vmware_rest/docs/*',
            'vmware/vmware_rest/tests/*',
            'vultr/cloud/.*',
            'vultr/cloud/tests/*',
            'vyos/vyos/.*',
            'vyos/vyos/docs/*',
            'vyos/vyos/tests/*',
            'wti/remote/docs/*',
        ],
    },
    include_package_data=True,
    install_requires=[
        'ansible-core ~= 2.14.5',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Framework :: Ansible',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
    data_files=[],
    entry_points={
        'console_scripts': [
            'ansible-community=ansible_collections.ansible_community:main'
        ]
    },
    # Installing as zip files would break due to references to __file__
    zip_safe=False
)


# Copyright (c) Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible.module_utils.common.text.converters import to_native


def parse_deps(content):
    decoded_contents = to_native(content)
    lines = [
        line.strip() for line in decoded_contents.split('\n')
        if line.strip() and not line.strip().startswith('#')
    ]
    deps = {}
    for line in lines:
        record = [entry.strip() for entry in line.split(':', 1)]
        deps[record[0]] = record[1]
    return deps


class FilterModule(object):
    ''' Deps file parsing filters '''

    def filters(self):
        filters = {
            '_antsibull_parse_deps': parse_deps,
        }

        return filters



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2023
"""Helper to create a Galaxy context."""

from __future__ import annotations

import aiohttp
from antsibull_core.galaxy import GalaxyContext


async def create_galaxy_context() -> GalaxyContext:
    """
    Create a Galaxy context.
    """
    async with aiohttp.ClientSession() as aio_session:
        return await GalaxyContext.create(aio_session)



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020

"""
Tools for building the Ansible Distribution
"""

from __future__ import annotations

__version__ = "0.60.0"



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Utility functions for dealing with paths and directories
"""

from __future__ import annotations

import contextlib
import tempfile
from collections.abc import AsyncIterator, Iterator
from pathlib import Path
from typing import TYPE_CHECKING

import aiofiles.ospath
import aiofiles.tempfile

from antsibull._vendor.shutil import copytree_and_symlinks

if TYPE_CHECKING:
    from _typeshed import StrPath


def _is_dir(directory: Path) -> None:
    if not directory.is_dir():
        raise ValueError(f"{directory} is not a directory!")


@contextlib.contextmanager
def temp_or_dir(directory: StrPath | None = None, /) -> Iterator[Path]:
    if directory:
        directory = Path(directory)
        _is_dir(directory)
        yield directory
    else:
        with tempfile.TemporaryDirectory() as tmp:
            yield Path(tmp)


@contextlib.asynccontextmanager
async def atemp_or_dir(directory: StrPath | None = None, /) -> AsyncIterator[Path]:
    if directory:
        directory = Path(directory)
        _is_dir(directory)
        yield directory
    else:
        async with aiofiles.tempfile.TemporaryDirectory() as tmp:
            yield Path(tmp)


__all__ = ("temp_or_dir", "atemp_or_dir", "copytree_and_symlinks")
